from easy_spider.network.request import Request
from easy_spider.network.client import AsyncClient
from easy_spider.network.response import Response, HTMLResponse
from easy_spider.extractors.extractor import SimpleBSExtractor
from easy_spider.core.recoverable import Recoverable
from easy_spider.error.error_handler import forward, handler_error
from easy_spider.middlewares.build_in import ChainMiddleware, ExtractorFilterMiddleware, FilterMiddleware, \
    GenerationMiddleware, SetAttrMiddleware
from easy_spider import tool
from abc import ABC, abstractmethod
from easy_spider.filters.build_in import Filter, html_filter, all_pass_filter, HashFilter, CrawledFilter
from typing import Iterable
from os.path import join


class Spider(ABC):

    def __init__(self):
        self._start_targets = []
        self.num_threads = 1
        self.num_of_spill = 10000
        self._filter = html_filter
        self.extractor = SimpleBSExtractor()
        self._crawled_filter = HashFilter()
        self.init()
        self._middlewares = self.middlewares()
        self._start_targets = list(self._middlewares.transform(self._start_targets, None))

    def init(self):
        pass

    def middlewares(self):
        return ChainMiddleware(GenerationMiddleware(),  # 设置请求为第 n 代
                               ExtractorFilterMiddleware(self.filter),  # 请求提取过滤器
                               FilterMiddleware(self.crawled_filter),  # 历史请求过滤器
                               SetAttrMiddleware(self))  # 设置请求默认参数

    @property
    def start_targets(self):
        return self._start_targets

    @start_targets.setter
    def start_targets(self, targets):
        self._start_targets = list(self.from_url_or_request_iter(targets))

    @property
    def crawled_filter(self):
        return self._crawled_filter

    @crawled_filter.setter
    def crawled_filter(self, filter):
        if filter and not isinstance(filter, CrawledFilter):
            raise TypeError("crawled_filter must be a CrawledFilter, got a {}".format(filter.__class__.__name__))
        self._crawled_filter = filter

    @property
    def filter(self): return self._filter

    @filter.setter
    def filter(self, filter):
        self._filter = filter or all_pass_filter

    @staticmethod
    def from_url_or_request(url: str):
        return Request.of(url)

    def from_url_or_request_iter(self, urls: Iterable[str]):
        yield from (self.from_url_or_request(url) for url in urls)

    @staticmethod
    def _nothing():
        return range(0)

    @abstractmethod
    def crawl(self, request: Request):
        pass


# class MultiThreadSpider(Spider, SimpleClient):
#
#     def __init__(self, handlers):
#         super().__init__(handlers)
#         self.start_requests = []
#
#     def crawl(self, request: Request):
#         response = self.do_request(request)  # 发送请求
#         return request.handler(response)


class AsyncSpider(Spider, AsyncClient):

    def __init__(self):
        super().__init__()

    @abstractmethod
    def handle(self, response: Response):
        """
        从 Response 中提取 Request，提取的 Request 将会设置默认参数
        """
        if isinstance(response, HTMLResponse):
            yield from self.from_url_or_request_iter(self.extractor.extract(response))
        else:
            yield from self._nothing()

    async def crawl(self, request: Request):
        """
        爬虫主要逻辑
        """
        response = await self.do_request(request)
        handler = forward(to=handler_error)(request.handler or self.handle)  # 包装错误处理
        request_like_iter = handler(response)
        if not request_like_iter:
            return self._nothing()
        return self._middlewares.transform(request_like_iter, response)


class RecoverableSpider(AsyncSpider, Recoverable, ABC):

    def __init__(self):
        self.name = None
        self.auto_save_frequency = 1000
        super().__init__()

    def stash(self, resource):
        self._crawled_filter.stash(resource)

    def recover(self, resource):
        self._crawled_filter.recover(resource)
        self._middlewares = self.middlewares()

    def can_recover(self, resource):
        return self._crawled_filter.can_recover(resource)



from abc import abstractmethod, ABC
from typing import Generator
from easy_spider.network.response import HTMLResponse


class Extractor(ABC):

    @abstractmethod
    def extract(self, response) -> Generator[str, None, None]: pass


class SimpleBSExtractor(Extractor):

    def extract(self, response: HTMLResponse) -> Generator[str, None, None]:
        for tag_a in response.bs.find_all("a"):
            href = tag_a.get("href")
            if href:
                yield response.url_join(href)




import unittest
from easy_spider.network.request import Request
from easy_spider.core.env import async_env
from easy_spider.core.spider import AsyncSpider, RecoverableSpider
from easy_spider.filters.build_in import GenerationFilter, URLRegFilter
# from test import mock_env


class MySpider(AsyncSpider):

    def init(self):
        self.num_threads = 4
        self.start_targets = ["http://localhost:5000/test_extract"]

    def handle(self, response):
        print(response.bs.title)
        yield from super().handle(response)


class MyRecoverableSpider(RecoverableSpider):
    def init(self):
        self.num_threads = 1
        self.start_targets = ["https://github.blog/"]
        self.filter = URLRegFilter(r"^https://github\.blog/page/\d+/$")
        self.num_threads = 1

    def handle(self, response):
        print(response.bs.title)
        yield from super().handle(response)


class TestCore(unittest.TestCase):

    def setUp(self) -> None:
        self.my_spider = MySpider()

    # def test_set_default_method(self):
    #     r = Request("test")
    #     self.my_spider.method = 'POST'
    #     self.my_spider._set_default_request_param(r)
    #     self.assertEqual(r.method, 'POST')
    #
    # def test_core(self):
    #     self.my_spider.method = 'GET'
    #     async_env.run(self.my_spider)
    #
    # def test_spider_with_generation_filter(self):
    #     spider = MySpider()
    #     spider.filter = spider.filter + GenerationFilter(max_generation=1)
    #     async_env.run(spider)

    def test_recover(self):
        spider = MyRecoverableSpider()
        with self.assertRaises(ValueError):
            async_env.run(spider)
        spider.name = "test_recover"
        async_env.run(spider)


if __name__ == '__main__':
    unittest.main()



from easy_spider.filters.build_in import *
from easy_spider.network.request import Request
import unittest


class TestFilter(unittest.TestCase):

    def test_reg_filter(self):
        f = URLRegFilter(r"http://")
        self.assertTrue(f.accept(Request.of("http://www.baidu.com")))

    def assert_true(self, r, url):
        self.assertTrue(r.accept(Request.of(url)))

    def assert_false(self, r, url):
        self.assertFalse(r.accept(Request.of(url)))

    def test_and_op(self):
        r = URLRegFilter("") + URLRegFilter("") + URLRegFilter("")
        self.assertEqual(type(r), AndChainFilter)

    def test_or_op(self):
        r = URLRegFilter("") | URLRegFilter("") | URLRegFilter("")
        self.assertEqual(type(r), OrChainFilter)

    def test_and_filter(self):
        r = URLRegFilter("http://") + URLRegFilter(".*?test")
        self.assert_false(r, "http://www.baidu.com")
        self.assert_true(r, "http://www.test.com")

    def test_or_filter(self):
        r = URLRegFilter("http://") | URLRegFilter("https://") + URLRegFilter(".*jpg$")
        self.assert_true(r, "http://www.baidu.com")
        self.assert_false(r, "https://www.baidu.com")
        self.assert_true(r, "https://www.baidu.com.jpg")

        r = (URLRegFilter("a") | URLRegFilter("b")) + (URLRegFilter("a") | URLRegFilter("c"))
        self.assert_true(r, "a")
        self.assert_false(r, "b")
        self.assert_false(r, "c")
        self.assert_false(r, "d")

        r = (URLRegFilter("^a") + URLRegFilter(".*b$")) | (URLRegFilter("^c") + URLRegFilter(".*d$"))
        self.assert_true(r, "ab")
        self.assert_true(r, "cd")
        self.assert_false(r, "c")
        self.assert_false(r, "d")

    def test_not_filter(self):
        r = -URLRegFilter("javascript:")
        self.assert_false(r, "javascript:")

    def test_and_not_filter(self):
        r = URLRegFilter(".*") + (-URLRegFilter("javascript:"))
        self.assert_true(r, "http://www.baidu.com")
        self.assert_false(r, "javascript:func(1)")

        r = URLRegFilter(".*") - URLRegFilter("javascript:") - URLRegFilter("https://")
        self.assert_true(r, "http://www.baidu.com")
        self.assert_false(r, "https://www.baidu.com")
        self.assert_false(r, "javascript:func(1)")

    def test_html_filter(self):
        self.assert_true(html_filter, "http://www.baidu.com")
        self.assert_true(static_filter, "http://www.baidu.com/a.zip")
        self.assert_false(html_filter, "http://www.baidu.com/a.zip")
        self.assert_false(html_filter, "http://www.baidu.com/b.Mp4")
        self.assert_false(html_filter, "http://www.baidu.com/b.mP3")

    def test_history_filter(self):
        hf = BloomFilter(1000, 0.001)
        self.assert_true(hf, "http://123123")
        self.assert_false(hf, "http://123123")

        hf = HashFilter()
        self.assert_true(hf, "http://123123")
        self.assert_false(hf, "http://123123")


if __name__ == '__main__':
    unittest.main()



from abc import ABC, abstractmethod
from easy_spider.network.request import Request
from easy_spider.network.response import Response
from easy_spider.tool import get_public_attr, copy_attr
from typing import List, Iterable, Optional
from easy_spider.filters.build_in import Filter


class RequestMiddleware(ABC):

    @abstractmethod
    def transform(self, requests: Iterable[Request], response: Optional[Response]) -> Iterable[Request]: pass


class FilterMiddleware(RequestMiddleware):
    def __init__(self, filter: Filter):
        self._filter = filter

    def transform(self, requests: Iterable[Request], response: Optional[Response]) -> Iterable[Request]:
        yield from filter(self._filter.accept, requests)


class ExtractorFilterMiddleware(FilterMiddleware):
    def __init__(self, filter: Filter):
        super().__init__(filter)

    def transform(self, requests: Iterable[Request], response: Optional[Response]) -> Iterable[Request]:
        if not response:  # 如果 response 为空，则说明为初始 requests，不需要进行过滤
            yield from requests
        else:
            yield from super().transform(requests, response)


class GenerationMiddleware(RequestMiddleware):

    def transform(self, requests: Iterable[Request], response: Optional[Response]) -> Iterable[Request]:
        if response:
            source_request = response.request
            for request in requests:
                request.generation = source_request.generation + 1
                yield request
        else:
            yield from requests


class SetAttrMiddleware(RequestMiddleware):

    def __init__(self, from_instance, attrs=None):
        self._from_instance = from_instance
        self._attrs = attrs

    def transform(self, requests: Iterable[Request], response: Optional[Response]) -> Iterable[Request]:
        for request in requests:
            for attr in (self._attrs or get_public_attr(request)):  # 如果有指定的 attr 则使用指定的 attr, 否则动态获取
                hasattr(self._from_instance, attr) and copy_attr(attr, self._from_instance, request)
            yield request


class ChainMiddleware(RequestMiddleware):

    def __init__(self, *middlewares: RequestMiddleware):
        self._middlewares: List[RequestMiddleware] = list(middlewares)

    @property
    def middlewares(self): return self._middlewares

    def transform(self, requests: Iterable[Request], response: Optional[Response]) -> Iterable[Request]:
        r = requests
        for middleware in self._middlewares:
            r = middleware.transform(r, response)
        yield from r

    def extend(self, middleware: 'ChainMiddleware'):
        self._middlewares.extend(middleware.middlewares)
        return self




class KnownError(Exception):
    error_type = ""


class ClientError(KnownError):
    error_type = "请求错误"


class ConnectionError(ClientError):
    error_type = "连接错误"


class ResponseError(ClientError):

    def __init__(self, code):
        self.code: int = code

    error_type = "响应错误"


class HandlerError(KnownError):
    error_type = "handler错误"



import unittest
from easy_spider.filters.build_in import *
from easy_spider.middlewares.build_in import *


class A:
    def __init__(self):
        self.method = "POST"


class MockResponse:
    def __init__(self):
        self.request = Request("test", generation=0)


class TestMiddlewares(unittest.TestCase):

    def test_middlewares(self):
        requests = [Request.of("http://www.test"), Request.of("http://a.test")]
        filter_mw = ExtractorFilterMiddleware(URLRegFilter("^https?://www\.test"))
        history_wm = FilterMiddleware(HashFilter())
        copy_wm = SetAttrMiddleware(A())
        gen_wm = GenerationMiddleware()
        wms = ChainMiddleware(gen_wm, filter_mw, history_wm, copy_wm)
        requests_transformed = list(wms.transform(requests, None))
        self.assertEqual(len(list(requests_transformed)), 2)
        self.assertTrue(all(map(lambda r: r.generation == 0, requests_transformed)))
        self.assertTrue(all(map(lambda r: r.method == "POST", requests_transformed)))
        requests_transformed = list(wms.transform(requests, None))
        self.assertEqual(len(list(requests_transformed)), 0)

        filter_mw = ExtractorFilterMiddleware(URLRegFilter("^https?://www\.test"))
        history_wm = FilterMiddleware(HashFilter())
        copy_wm = SetAttrMiddleware(A())
        gen_wm = GenerationMiddleware()
        wms = ChainMiddleware(gen_wm, filter_mw, history_wm, copy_wm)
        requests_transformed = list(wms.transform(requests, MockResponse()))
        self.assertEqual(len(requests_transformed), 1)
        self.assertTrue(all(map(lambda r: r.generation == 1, requests_transformed)))
        self.assertTrue(all(map(lambda r: r.method == "POST", requests_transformed)))


if __name__ == "__main__":
    unittest.main()



from abc import ABC, abstractmethod
from easy_spider.network.request import Request
from easy_spider.network.response import (Response, TextResponse, HTMLResponse)
from easy_spider.error.error_handler import client_error
from easy_spider.error.known_error import ResponseError
from aiohttp import ClientSession, ClientTimeout


class Client(ABC):

    @abstractmethod
    def do_request(self, resource: Request) -> Response: pass

    @staticmethod
    def to_response(request, content, url, headers):
        content_type = headers.get("Content-Type")
        args = (content, url, headers)
        response = None
        if content_type:
            if "text/html" in content_type:
                response = HTMLResponse(*args)
            elif "text/" in content_type:
                response = TextResponse(*args)
        else:
            response = Response(*args)
        response.request = request
        return response


class AsyncClient(Client):

    def __init__(self, session=None):
        self._session: ClientSession = session

    def set_session(self, session):
        self._session = session

    async def do_request(self, request: Request):
        assert request
        common_params = dict(
            headers=request.headers,
            cookies=request.cookies,
            timeout=ClientTimeout(request.timeout),
            params=request.params,
            proxy=request.proxy
        )
        data_format = request.data_format.lower()
        if data_format == "form":
            common_params["data"] = request.data
        elif data_format == "json":
            common_params["json"] = request.data
        else:
            raise ValueError("未知数据类型 {} ，仅支持 json 或 form")
        try:
            raw_response = await self._session.request(request.method, request.url, **common_params)
            content = await raw_response.content.read()
        except Exception as e:
            raise client_error(e) from e
        if raw_response.status >= 400:
            raise ResponseError(raw_response.status) from None
        return self.to_response(request, content, str(raw_response.url), raw_response.headers)



from easy_spider.filters.filter import *
from easy_spider.tool import get_extension
from easy_spider.filters.history_filter import *

# 非 html 后缀， 来源于 scrapy
# https://github.com/scrapy/scrapy/blob/master/scrapy/linkextractors/__init__.py
_IGNORED_EXTENSIONS = {
    # archives
    '7z', '7zip', 'bz2', 'rar', 'tar', 'tar.gz', 'xz', 'zip',

    # images
    'mng', 'pct', 'bmp', 'gif', 'jpg', 'jpeg', 'png', 'pst', 'psp', 'tif',
    'tiff', 'ai', 'drw', 'dxf', 'eps', 'ps', 'svg', 'cdr', 'ico',

    # audio
    'mp3', 'wma', 'ogg', 'wav', 'ra', 'aac', 'mid', 'au', 'aiff',

    # video
    '3gp', 'asf', 'asx', 'avi', 'mov', 'mp4', 'mpg', 'qt', 'rm', 'swf', 'wmv',
    'm4a', 'm4v', 'flv', 'webm',

    # office suites
    'xls', 'xlsx', 'ppt', 'pptx', 'pps', 'doc', 'docx', 'odt', 'ods', 'odg',
    'odp',

    # other
    'css', 'pdf', 'exe', 'bin', 'rss', 'dmg', 'iso', 'apk', 'js'
}


def URLRegFilter(reg_expr):
    """
    使 RegFilter 直接作用于 Request 对象的 uri
    """
    return URLFilter(RegexFilter(reg_expr))


static_filter = URLFilter(CustomFilter(lambda url: get_extension(url) in _IGNORED_EXTENSIONS))
url_filter = URLRegFilter(r"^https?:\/{2}[^\s]*?(\?.*)?$")
html_filter = url_filter - static_filter
all_pass_filter = CustomFilter(lambda _: True)
all_reject_filter = CustomFilter(lambda _: False)


class GenerationFilter(Filter):
    def __init__(self, max_generation):
        self.max_generation = max_generation

    def accept(self, request: Request) -> bool:
        return request.generation <= self.max_generation



import unittest
from easy_spider.core.spider import AsyncSpider, RecoverableSpider
from easy_spider.filters.build_in import BloomFilter, all_pass_filter, CustomFilter, GenerationFilter
from easy_spider.network.request import Request
from test.mock_env import run_and_get_result, env
from easy_spider.tool import EXE_PATH
from os.path import join


class MySpider(AsyncSpider):

    def init(self):
        self.start_targets = ["http://localhost:5000/test_extract"]
        self.num_threads = 4

    def handle(self, response):
        print(response.bs.title)
        yield from super().handle(response)


class RecoverMySpider(RecoverableSpider):

    def init(self):
        self.start_targets = ["http://localhost:5000/test_extract"]
        self.num_threads = 4
        self.filter = self.filter + GenerationFilter(max_generation=1)

    def handle(self, response):
        print(response.bs.title)
        yield from super().handle(response)


class TestSpider(unittest.TestCase):

    async def async_spider(self):
        r = Request("http://localhost:5000/test_extract")
        spider = MySpider()
        spider.set_session(env.session)
        requests = await spider.crawl(r)
        for request in requests:
            print(request)

    def test_async_spider(self):
        run_and_get_result(self.async_spider())

    def test_crawled_spider(self):
        spider = MySpider()
        with self.assertRaises(TypeError):
            spider.crawled_filter = spider.filter
        start_request = Request.of("http://localhost:5000/test_extract")
        self.assertFalse(spider.crawled_filter.accept(start_request))

    def test_recover(self):
        spider = RecoverMySpider()
        spider.filter = spider.filter + CustomFilter(lambda x: True)
        spider.set_session(env.session)
        list(run_and_get_result(spider.crawl(Request.of("http://localhost:5000/test_extract"))))
        spider.stash(EXE_PATH)
        del spider
        recovered_spider = RecoverMySpider()
        recovered_spider.recover(EXE_PATH)
        self.assertTrue(recovered_spider.crawled_filter.contains(Request.of("http://localhost:5000/a.html")))


if __name__ == '__main__':
    unittest.main()



import unittest
from easy_spider.core.task import RecoverableTask, AsyncTask, CountDownRecoverableTask
from easy_spider.core.spider import AsyncSpider, RecoverableSpider
from easy_spider.network.request import Request
from os import makedirs
from os.path import exists
from test import mock_env


class MySpider(RecoverableSpider):
    def __init__(self):
        super().__init__()
        self.start_targets = ["http://www.test.com"]
        self.name = "test_spider"

    def handle(self, response):
        pass


class TestTask(unittest.TestCase):
    def test_stash(self):
        spider = MySpider()
        spider.start_targets = ["http://localhost:5000/test_extract"]
        task = RecoverableTask(spider)
        exists(".task_stash") or makedirs(".task_stash")
        task.stash(".task_stash")
        del task
        recovered_task = RecoverableTask(MySpider())
        self.assertTrue(recovered_task.can_recover(".task_stash"))
        recovered_task.recover(".task_stash")
        self.assertTrue(recovered_task.spider.crawled_filter.contains(Request.of("http://localhost:5000/test_extract")))

    def test_auto_save(self):
        test_data = {"key": 1}
        spider = MySpider()
        spider.set_session(mock_env.env.session)
        spider.start_targets = ["http://localhost:5000/test_extract"]
        spider.auto_save_frequency = 10
        task = CountDownRecoverableTask(spider)
        task.add_actions(lambda: test_data.clear())
        for _ in range(5):
            mock_env.env.loop.run_until_complete(task.run())
            task._request_queue.put(Request.of("http://localhost:5000/test_extract"))
        self.assertEqual(test_data, {"key": 1})
        for _ in range(5):
            mock_env.env.loop.run_until_complete(task.run())
            task._request_queue.put(Request.of("http://localhost:5000/test_extract"))
        self.assertEqual(test_data, {})


if __name__ == '__main__':
    unittest.main()



from easy_spider.core.spider import AsyncSpider, RecoverableSpider
from easy_spider.core.recoverable import Recoverable, CountDown, FileBasedRecoverable
from easy_spider.log import console_logger, file_logger
from easy_spider.core.queue import get_queue_for_spider, RequestQueue
from easy_spider.error.known_error import ClientError
from easy_spider.error.error_formatter import ErrorFormatter, DefaultErrorFormatter, ClientErrorFormatter
import asyncio
from abc import ABC, abstractmethod
from easy_spider.tool import EXE_PATH
from os.path import join, exists
from os import listdir
from typing import List


class Task(ABC):

    @abstractmethod
    def run(self): pass


class AbstractTask(Task, ABC):

    def __init__(self, spider, request_queue: RequestQueue):
        self._spider = spider
        self._request_queue = request_queue

    def _init_queue(self):
        if not self._spider.start_targets:
            raise ValueError("初始请求不能为空")
        self._request_queue.put_many(self._spider.start_targets)

    @property
    def spider(self): return self._spider


# class MultiThreadJob(AbstractTask):
#     def __init__(self, spider: MultiThreadSpider, request_queue):
#         super().__init__(spider, request_queue)
#         self._thread_pool = ThreadPoolExecutor(max_workers=spider.start_requests)
#         self._num_running_task = Value("i", 0)
#         self._lock = Lock()
#
#     def run(self):
#         fs = [self._thread_pool.submit(self._run) for _ in range(self._spider.num_threads)]
#         wait(fs)
#
#     def _run(self):
#         while True:
#             with self._lock:
#                 resource = self._request_queue.get()
#                 if resource is None:
#                     if self._num_running_task.value == 0:  # 如果 task_queue 为空, 且同时正在进行的任务为0则退出
#                         break
#                     else:
#                         sleep(0)  # 否则 sleep 等待任务
#                         continue
#                 else:
#                     self._num_running_task.value += 1
#             try:
#                 new_resources = self._spider.crawl(resource)
#                 self._request_queue.put_many(new_resources)
#             except Exception as e:
#                 logger.warning(f"{resource}处理失败: {e}", exc_info=True)
#             finally:
#                 self._num_running_task.value -= 1


class AsyncTask(AbstractTask):

    def __init__(self, spider: AsyncSpider, request_queue=None):
        request_queue = get_queue_for_spider(spider) if request_queue is None else request_queue
        super().__init__(spider, request_queue)
        self._init_queue()
        self._progress_requests = []  # 正在处理中的请求
        self._error_formatter = ErrorFormatter(DefaultErrorFormatter())
        self._error_formatter.register(ClientError, ClientErrorFormatter())

    async def run(self):
        await asyncio.gather(*[self._run() for _ in range(self._spider.num_threads)])

    async def _wait_request(self):
        while True:
            request = self._request_queue.get()
            # 当队列中没有 request 且正在处理的 request 不为空，则有可能会有新的request，因此应该等待
            # 否则直接返回 None 表示不会再有新的 request 产生
            if request is None and self._progress_requests:
                await asyncio.sleep(1)
                continue
            return request

    async def _crawl_request(self, request):
        self._progress_requests.append(request)
        try:
            new_requests = await self._spider.crawl(request)
            self._request_queue.put_many(new_requests)
            console_logger.info("成功 {} ".format(request))
        except Exception as e:
            console_logger.warning("失败 %s %s", request, self._error_formatter.format(e))
            file_logger.warning("失败 %s", request, exc_info=True)
        self._progress_requests.remove(request)

    async def _run(self):
        while True:
            request = await self._wait_request()
            if not request:
                break
            await self._crawl_request(request)
        console_logger.info("协程已退出")


class RecoverableTask(AsyncTask, FileBasedRecoverable):

    def __init__(self, spider: RecoverableSpider, request_queue=None):
        # 不能改为 request_queue = request_queue or get_queue_for_spider(spider) !!!
        request_queue = get_queue_for_spider(spider) if request_queue is None else request_queue
        super().__init__(spider, request_queue)
        self._recover_items = (spider, request_queue)
        FileBasedRecoverable.__init__(self)

    def stash_attr_names(self) -> List[str]:
        return ["_progress_requests"]

    def can_recover(self, resource):
        if not exists(resource) or not listdir(resource):
            return False
        for item in self._recover_items:
            if not item.can_recover(resource):
                raise FileNotFoundError("can't recover {}, the file is missing".format(item.recover_name()))
        return True

    def stash(self, resource):
        for recover_item in self._recover_items:
            recover_item.stash(resource)
        super().stash(resource)

    def recover(self, resource):
        if not self.can_recover(resource):
            raise ValueError("{} not exist".format(join(EXE_PATH, resource)))
        for recover_item in self._recover_items:
            recover_item.recover(resource)
        try:
            super().recover(resource)  # todo: 兼容以前的任务，待删除
        except:
            pass
        self._request_queue.put_many(self._progress_requests)  # 将未处理的 request 重新放入队列
        self._progress_requests.clear()  # 清空未处理队列


class CountDownRecoverableTask(RecoverableTask, CountDown):
    def __init__(self, spider: RecoverableSpider, request_queue=None):
        request_queue = request_queue or get_queue_for_spider(spider)
        RecoverableTask.__init__(self, spider, request_queue)
        CountDown.__init__(self, spider.auto_save_frequency)

    async def _crawl_request(self, request):
        await super()._crawl_request(request)
        self.count()



import unittest
from easy_spider.error.known_error import *
from asyncio import TimeoutError
from aiohttp.client_exceptions import ClientConnectionError
from easy_spider.error.error_formatter import ClientErrorFormatter, DefaultErrorFormatter, ErrorFormatter
from easy_spider.error.error_handler import handler_error, client_error, forward
from easy_spider.error.known_error import ClientError, ResponseError
from easy_spider.core.spider import AsyncSpider
from easy_spider import Request
from test import mock_env


@forward(to=handler_error)
def handler():
    raise ValueError("输入错误")


@forward(to=client_error)
def do_request():
    raise TimeoutError()


class Spider(AsyncSpider):
    def __init__(self):
        super().__init__()

    def handle(self, response):
        raise ValueError("输入错误")


class TestExceptionFormatter(unittest.TestCase):

    def test_formatter(self):
        formatter = ErrorFormatter(DefaultErrorFormatter())
        formatter.register(ClientError, ClientErrorFormatter())

        try:
            raise ConnectionError() from ClientConnectionError()
        except Exception as e:
            self.assertIn("aiohttp.client_exceptions.ClientConnectionError", ClientErrorFormatter().format(e))

        try:
            raise ConnectionError() from TimeoutError()
        except Exception as e:
            self.assertIn("超时", ClientErrorFormatter().format(e))

        try:
            raise HandlerError() from ValueError("输入错误")
        except Exception as e:
            self.assertIn("输入错误", DefaultErrorFormatter().format(e))

        try:
            raise ValueError("输入错误")
        except Exception as e:
            self.assertIn("输入错误", DefaultErrorFormatter().format(e))

    def test_handler(self):
        with self.assertRaises(HandlerError):
            handler()

        with self.assertRaises(ClientError):
            do_request()

    def test_spider_handler(self):
        spider = Spider()
        spider.set_session(mock_env.env.session)
        formatter = ErrorFormatter(DefaultErrorFormatter())
        formatter.register(ClientError, ClientErrorFormatter())
        with self.assertRaises(HandlerError):
            mock_env.run_and_get_result(spider.crawl(Request.of("http://localhost:5000/test_extract")))
        with self.assertRaises(ConnectionError):
            mock_env.run_and_get_result(spider.crawl(Request.of("no exist")))
        with self.assertRaises(ResponseError):
            mock_env.run_and_get_result(spider.crawl(Request.of("http://localhost:5000/error")))


if __name__ == "__main__":
    unittest.main()



from easy_spider.error.known_error import KnownError
from asyncio import TimeoutError
from abc import ABC, abstractmethod
from typing import Tuple, Optional


class ErrorFormatter:
    def __init__(self, default_formatter):
        self._default_formatter = default_formatter
        self._formatters = {}

    def register(self, exception_cls, formatter):
        self._formatters[exception_cls] = formatter

    def format(self, exception):
        formatter = self._formatters.get(exception.__class__, self._default_formatter)
        return formatter.format(exception)


class Formatter(ABC):

    @abstractmethod
    def _get_formatter_items(self, exception) -> Tuple[str, Optional[Exception]]: pass

    def format(self, exception):
        exception_type, cause = self._get_formatter_items(exception)
        if not str(cause):  # 如果 cause 为空，则将异常类的名字作为 cause
            cause = "{}.{}".format(cause.__module__, cause.__class__.__name__)
        if not hasattr(exception, "code"):
            return "{} -> {}".format(exception_type, cause)
        else:
            return "{} -> {} code: {}".format(exception_type, cause, exception.code)


class DefaultErrorFormatter(Formatter):

    def _get_formatter_items(self, exception) -> Tuple[str, Optional[Exception]]:
        if isinstance(exception, KnownError):
            return exception.error_type, exception.__cause__
        else:
            return "运行时错误", exception


class ClientErrorFormatter(DefaultErrorFormatter):

    def _get_formatter_items(self, exception) -> Tuple[str, Optional[Exception]]:
        if exception.__cause__.__class__ == TimeoutError:
            return exception.error_type, Exception("连接超时")
        else:
            return super()._get_formatter_items(exception)



from abc import ABC, abstractmethod
from os.path import exists, join
from typing import List
from easy_spider.tool import pickle_load, pickle_dump, get_type_name


class Recoverable(ABC):

    @abstractmethod
    def stash(self, resource): pass

    @abstractmethod
    def recover(self, resource): pass

    @classmethod
    def recover_name(cls): return cls.__name__

    @abstractmethod
    def can_recover(self, resource): pass


class FileBasedRecoverable(Recoverable, ABC):
    """
        利用 pickle 对爬虫进行保存
        子类只需要实现 stash_attr_names，给出需要保存的 attr
        对于需要保存的所有 attr，会形成字典 {attr1, value1, ..., } 并利用 pickle 保存
        恢复时使用 setattr 恢复
    """
    def __init__(self):
        self._stash_attr_names = self.stash_attr_names()
        for attr in self._stash_attr_names:
            self._check_stash_attr(attr)

    @classmethod
    def recover_name(cls):
        return super().recover_name() + ".pickle"

    def _check_stash_attr(self, attr):
        if attr not in self.__dict__:
            raise AttributeError("can't find stash attr `{}`".format(attr))

    @abstractmethod
    def stash_attr_names(self) -> List[str]: pass

    def can_recover(self, resource):
        return exists(self._get_stash_file_uri(resource))

    def _get_stash_file_uri(self, resource):
        return join(resource, self.recover_name())

    def stash(self, resource):
        """
        将需要保存的 attr 存放到 {resource}/{self.__class__.__name__}.pickle
        :param resource: 存放文件路径
        :return: None
        """
        attrs_to_stash = {attr: value for attr, value in self.__dict__.items() if attr in self._stash_attr_names}
        pickle_dump(attrs_to_stash, self._get_stash_file_uri(resource))

    def recover(self, resource):
        """
        从 {resource}/{self.__class__.__name__}.pickle 中恢复
        :param resource: 存放文件路径
        :return: None
        """
        attrs_to_stash = pickle_load(self._get_stash_file_uri(resource))
        for attr, value in attrs_to_stash.items():
            setattr(self, attr, value)

# class BufferedRecoverable(ABC):
#
#     @abstractmethod
#     def write(self): pass
#
#     @abstractmethod
#     def recover(self): pass


class CountDown:
    def __init__(self, start):
        if start <= 0:
            raise ValueError("start must > 1, get `{}`".format(start))
        self._start = start
        self._now = start
        self._actions = []

    def add_actions(self, action):
        if not callable(action):
            raise TypeError("action must be a callable, got `{}`".format(get_type_name(action)))
        self._actions.append(action)

    def count(self):
        self._now -= 1
        if self._now == 0:
            for action in self._actions:
                action()
            self._now = self._start



from easy_spider.core.env import async_env
from easy_spider.core.spider import AsyncSpider, RecoverableSpider
from easy_spider.network.request import Request
from easy_spider.network.response import Response, HTMLResponse, TextResponse
from easy_spider.filters.build_in import *

__version__ = "1.0.14"
name = "easy_spider"



import posixpath
from typing import Any
from os import path, remove
import sys
import pickle
from uuid import uuid1
from datetime import datetime


def get_extension(url):
    return posixpath.splitext(url)[-1].lower().lstrip(".")


def get_public_attr(instance: Any):
    for attr, value in instance.__dict__.items():
        if not attr.startswith("_"):
            yield attr


def copy_attr(attr, _from, _to):
    setattr(_to, attr, getattr(_from, attr))


def get_abs_path(file_path):
    return path.dirname(path.abspath(file_path))


def pickle_dump(obj, file_uri):
    with open(file_uri, 'wb') as wd:
        pickle.dump(obj, wd)


def pickle_load(file_uri):
    with open(file_uri, 'rb') as rd:
        return pickle.load(rd)


def uuid():
    return str(uuid1())


def formatted_datetime(fmt=None):
    return datetime.utcnow().strftime(fmt or '%Y-%m-%d %H:%M:%S.%f')


def delete_file(uri):
    try:
        remove(uri)
    except IOError as e:
        # console_logger.warning("delete spill file %s error" % uri)
        # file_logger.warning("delete spill file %s error" % uri, exc_info=True)
        pass


def get_type_name(instance):
    return instance.__class__.__name__


ABS_PATH = get_abs_path(__file__)
sys.path.append(path.join(ABS_PATH, ".."))
EXE_PATH = path.dirname(path.abspath(__name__))


def work_path_join(arg, *args):
    return path.join(EXE_PATH, '.easy-spider', arg, *args)


def confirm(info):
    return input("[*] {}? (y/n): ".format(info)).lower() == "y"



from easy_spider.extractors.extractor import *
import unittest
from test.mock_env import get


class Extractor(unittest.TestCase):

    @staticmethod
    def extract_urls(url):
        extractor = SimpleBSExtractor()
        return list(extractor.extract(get(url)))

    def test_simple_extractor(self):
        urls = self.extract_urls("http://localhost:5000/test_extract")
        self.assertIn("http://localhost:5000/a.html", urls)
        self.assertIn("http://localhost:5000/b/c/d", urls)
        self.assertIn("javascript:func(1)", urls)
        self.assertIn("http://localhost:5000/test.zip", urls)
        self.assertIn("http://localhost:5000/test.Mp4", urls)
        self.assertIn("http://localhost:5000/test.mP3", urls)


if __name__ == "__main__":
    unittest.main()



from abc import ABC, abstractmethod
from collections import deque
from easy_spider.core.recoverable import FileBasedRecoverable
from easy_spider.tool import EXE_PATH, uuid, pickle_load, pickle_dump, formatted_datetime, delete_file, work_path_join
from os.path import join, exists
from os import makedirs
from easy_spider.log import console_logger
from typing import List


class Request(ABC):
    def __init__(self, url, handler=None, method='GET', priority=0, tag="any",
                 headers=None, cookies=None, timeout=None, params=None, data=None,
                 data_format='form', encoding='utf-8', generation=0, proxy=None):
        self.url: str = url
        self._handler = handler
        self.priority: int = priority
        self.tag = tag
        self.method: str = method
        self.timeout: int = timeout
        self.headers: dict = headers or {}
        self.cookies: dict = cookies or {}
        self.encoding: str = encoding
        self.params: dict = params or {}
        self.data: dict = data
        self.data_format: str = data_format
        self.proxy: str = proxy
        self.generation = generation

    @property
    def handler(self):
        return self._handler

    @handler.setter
    def handler(self, callback):
        if not callable(callback):
            raise TypeError("handler must be a callable, but get type: `{}`".format(type(callback).__name__))
        self._handler = callback

    def __repr__(self):
        return "[method={} uri={}]".format(self.method, self.url)

    def __str__(self):
        return self.__repr__()

    @staticmethod
    def of(instance):
        if isinstance(instance, Request):
            return instance
        elif isinstance(instance, str):
            return Request(instance)
        else:
            raise TypeError("can't build request from type `{}`".format(type(instance).__name__))


from abc import ABC, abstractmethod
from easy_spider.filters.filter import DependenceFilter
from typing import Any, List
from easy_spider.network.request import Request
from bloom_filter import BloomFilter as _BloomFilter
from easy_spider.core.recoverable import FileBasedRecoverable


class CrawledFilter(ABC):
    """
        已爬取过滤器， 依赖于其他过滤器的结果
        pre_filter: 前置过滤器，已爬取过滤器将依赖于前置过滤器的返回结果
        若已爬取过滤器以及前置过滤器都返回 True 则记录到已爬取过滤器，并返回 True
        否则返回 False
    """

    @abstractmethod
    def contains(self, request: Request) -> bool: pass

    @abstractmethod
    def add(self, request: Request): pass

    def accept(self, request: Request) -> bool:
        history_filter_accept = not self.contains(request)
        history_filter_accept and self.add(request)  # 如果不存在于布隆过滤器中，则记录
        return history_filter_accept

    @abstractmethod
    def clear(self): pass


class BloomFilter(CrawledFilter, FileBasedRecoverable):

    def stash_attr_names(self) -> List[str]:
        return ["_history_filter"]

    def __init__(self, max_elements, error_rate):
        CrawledFilter.__init__(self)
        self._max_elements = max_elements
        self._error_rate = error_rate
        self._history_filter = _BloomFilter(max_elements, error_rate)
        FileBasedRecoverable.__init__(self)

    def contains(self, request: Request):
        return request.url in self._history_filter

    def add(self, request: Request):
        self._history_filter.add(request.url)

    def clear(self):
        self._history_filter = _BloomFilter(self._max_elements, self._error_rate)


class HashFilter(CrawledFilter, FileBasedRecoverable):

    def __init__(self):
        self._crawled_urls = set()
        CrawledFilter.__init__(self)
        FileBasedRecoverable.__init__(self)

    def stash_attr_names(self) -> List[str]:
        return ["_crawled_urls"]

    def contains(self, request: Request) -> bool:
        return request.url in self._crawled_urls

    def add(self, request: Request):
        self._crawled_urls.add(request.url)

    def clear(self):
        self._crawled_urls.clear()



from easy_spider.network.client import AsyncClient
from easy_spider.core.task import AsyncTask, RecoverableTask, CountDownRecoverableTask
from easy_spider.core.spider import AsyncSpider, RecoverableSpider
from easy_spider.log import console_logger
from easy_spider.tool import EXE_PATH, work_path_join, confirm
from aiohttp import ClientSession
from asyncio import get_event_loop
from cached_property import cached_property
from os.path import join
from os import makedirs
from os.path import exists
import signal


class AsyncSpiderEvn:
    def __init__(self, loop=None):
        self._client = None
        self._loop = loop
        self._session = None

    def _prepare(self):
        if self._session is None:
            self._session = ClientSession()
        if self._client is None:
            self._client = AsyncClient()

    @staticmethod
    def _when_interrupt(_, __, task):
        if not confirm("You entered Ctrl + C, would you like stop spider"):
            return
        if not confirm("Would you like stash your spider task"):
            exit()
        spider_name = task.spider.name
        stash_path = work_path_join(spider_name)
        if task.can_recover(stash_path):
            if not confirm("The stash file of spider `{}` already exist, overwrite it".format(spider_name)):
                exit()
        print("[*] Start stash for spider `{}`".format(task.spider.name))
        task.stash(stash_path)
        print("[*] Stash success, exit")
        exit()

    @staticmethod
    def _recover_or_create(spider):
        if not hasattr(spider, "name") or not spider.name:
            raise ValueError("RecoverSpider must have property `name`")
        # 存储爬虫数据的文件夹为 EXE_PATH/.xxx, xxx=spider.name, EXE_PATH 为运行命令执行的文件夹
        stash_path = work_path_join(spider.name)
        if spider.auto_save_frequency != -1:
            task = CountDownRecoverableTask(spider)
            task.add_actions(lambda: task.stash(stash_path))  # 自动保存频率不为 1，则添加自动保存回调函数
        else:
            task = RecoverableTask(spider)
        if task.can_recover(stash_path):  # 如果能从文件中恢复则先进行恢复
            if confirm("Detected stashed spider, would you like to recover it"):
                print("[*] recover from exist stash file")
                task.recover(stash_path)
        exists(stash_path) or makedirs(stash_path)
        return task

    async def _run_spider(self, spider):
        self._prepare()
        spider.set_session(self._session)
        if isinstance(spider, RecoverableSpider):
            task = self._recover_or_create(spider)

            def callback_when_interrupt(signum, frame):
                self._when_interrupt(signum, frame, task)

            signal.signal(signal.SIGINT, callback_when_interrupt)
            signal.signal(signal.SIGTERM, callback_when_interrupt)
        else:
            task = AsyncTask(spider)
        await task.run()

    @cached_property
    def loop(self):
        if not self._loop:
            self._loop = get_event_loop()
        return self._loop

    def run(self, spider: AsyncSpider):
        self.loop.run_until_complete(self._run_spider(spider))

    def clear(self):
        self.__del__()

    def __del__(self):
        if not self._loop or self.loop.is_closed():
            return
        self._session and self.loop.run_until_complete(self._session.close())


async_env = AsyncSpiderEvn()



from aiohttp.client_exceptions import ClientConnectionError, ClientResponseError
from asyncio import TimeoutError
from easy_spider.error import known_error
from functools import wraps


def client_error(exception):
    if isinstance(exception, TimeoutError) or isinstance(exception, ClientConnectionError):
        return known_error.ConnectionError()
    elif isinstance(exception, ClientResponseError):
        return known_error.ResponseError(exception.status)
    else:
        return known_error.ClientError()


def handler_error(_):
    return known_error.HandlerError()


def forward(to):
    def inner(method):
        @wraps(method)
        def wrapper(*args, **kwargs):
            try:
                res = method(*args, **kwargs)
            except Exception as e:
                new_exception = to(e)
                raise new_exception from e
            return res
        return wrapper
    return inner



from easy_spider.network.request import Request
from easy_spider.core.spider import AsyncSpider, RecoverableSpider, Spider
from easy_spider.core.recoverable import FileBasedRecoverable
from abc import ABC, abstractmethod
from collections import deque
from os.path import join, exists
from os import makedirs
from easy_spider.tool import *
from easy_spider.log import console_logger
from typing import List


class RequestQueue(ABC):

    @abstractmethod
    def put(self, request: Request) -> None: pass

    def put_many(self, requests):
        for request in requests:
            self.put(request)

    @abstractmethod
    def get(self) -> Request: pass

    @abstractmethod
    def empty(self) -> bool: pass

    @abstractmethod
    def __len__(self): pass

    def __bool__(self):
        """
        Python 中逻辑值检测遵循以下准则:
        一个对象在默认情况下均被视为真值，除非当该对象被调用时其所属类定义了 __bool__() 方法且返回 False 或是定义了 __len__() 方法且返回零
        RequestQueue 所有子类均实现了 __len__ 方法。
        但是当:
            def a(request=None):
                request_b = request or SimpleRequestQueue()
            request_a = SimpleRequestQueue()
            a(request_a)
        时，request_a.__len__() 返回 0，导致 request_a or Request() 逻辑检测中 request_a 为假值，此时会生成新的对象。
        而非理想中的 request_b is request_a。
        """
        return len(self) == 0


# class SyncRequestQueue(RequestQueue):
#
#     def __init__(self):
#         super().__init__()
#         self._queue = Queue()
#
#     def put(self, request: Request) -> None:
#         self._queue.put(request)
#
#     def get(self):
#         try:
#             return self._queue.get_nowait()
#         except Empty:
#             return None
#
#     def empty(self) -> bool:
#         return self._queue.empty()
#
#     def __len__(self):
#         return self._queue.qsize()


class SimpleRequestQueue(RequestQueue):

    def __init__(self):
        self._queue = deque()

    def put(self, request: Request) -> None:
        self._queue.append(request)

    def get(self) -> Request:
        try:
            return self._queue.popleft()
        except IndexError:
            return None

    def clear(self):
        self._queue.clear()

    def empty(self) -> bool:
        return len(self) == 0

    def __len__(self):
        return len(self._queue)


class RecoverableRequestQueue(SimpleRequestQueue, FileBasedRecoverable, ABC):

    def __init__(self):
        SimpleRequestQueue.__init__(self)
        FileBasedRecoverable.__init__(self)

    def stash_attr_names(self):
        return ["_queue"]


class SpillRequestQueueProxy(RequestQueue):
    """
        代理类，对 RequestQueue 进行代理，增加在一定条件下将内存中的 request 写入磁盘以节约内存
        当 RequestQueue 为空时，从磁盘加载 requests。
        queue: 需要代理的 RequestQueue
        num_of_spill: 缓存数量
        写入策略: 当 RequestQueue 中 requests 数量 >= num_of_spill 时，将 requests 放入 self._wait_spill 中
        若 len(self._wait_spill) == num_of_spill，则将 self._spilled_files 写入磁盘文件。并将文件名放入 self._spilled_files
        加载策略: 当 RequestQueue 为空时，优先从 self._spilled_files 选取第一个文件加载到 RequestQueue
        否则从 self._wait_spill 中加载。
    """

    def __init__(self, queue: RequestQueue, num_of_spill=10000):
        self._queue = queue
        self._num_of_spill = num_of_spill
        self._spill_dir_name = 'spill-{}-{}'.format(formatted_datetime('%Y-%m-%d-%H-%M-%S-%f'), uuid())
        self._spill_path = work_path_join(self._spill_dir_name)
        self._wait_spill = []
        self._spilled_files = []

    def empty(self) -> bool:
        return len(self) == 0

    def _join(self, file_name):
        return join(self._spill_path, file_name)

    def _may_spill(self):
        """
        满足条件时，将 requests 写入磁盘
        """
        if len(self._wait_spill) == self._num_of_spill:
            spill_filename = uuid()
            exists(self._spill_path) or makedirs(self._spill_path)
            pickle_dump(self._wait_spill, self._join(spill_filename))  # spill 到文件
            self._spilled_files.append(spill_filename)  # 记录已经 spill 到磁盘的文件
            self._wait_spill.clear()

    def _may_load(self):
        """
        代理队列为空，则从磁盘加载 requests。优先从已经 spill 到磁盘的文件中加载，如果已经步不存在 spill 到磁盘的文件
        则从 self._wait_spill 中加载
        """
        if self._queue.empty():
            if self._spilled_files:
                spill_file_uri = self._join(self._spilled_files.pop(0))
                try:
                    self.put_many(pickle_load(spill_file_uri))
                except IOError:
                    console_logger.warning(
                        "can't load requests from file `%s`, that's mean you may lost some requests unhandled" % spill_file_uri)
                delete_file(spill_file_uri)  # 加载完成后删除磁盘上的文件
            elif self._wait_spill:
                self.put_many(self._wait_spill)
                self._wait_spill.clear()

    def __len__(self):
        return len(self._queue) + len(self._wait_spill) + len(self._spilled_files) * self._num_of_spill

    def get(self) -> Request:
        self._may_load()
        return self._queue.get()

    def put(self, request: Request) -> None:
        if len(self._queue) >= self._num_of_spill:  # 当代理 queue 中 request 的数量 = self._num_of_spill 时
            self._wait_spill.append(request)  # 新的 request 放入 self._wait_spill 中
            self._may_spill()
        else:
            self._queue.put(request)

    def put_many(self, requests):
        for request in requests:
            self.put(request)


class RecoverableSpillRequestQueueProxy(SpillRequestQueueProxy, FileBasedRecoverable):

    def __init__(self, queue: RecoverableRequestQueue, num_of_spill=10000):
        SpillRequestQueueProxy.__init__(self, queue, num_of_spill)
        FileBasedRecoverable.__init__(self)

    def stash_attr_names(self) -> List[str]:
        return ["_spill_dir_name", "_spill_path", "_spilled_files", "_wait_spill"]

    def stash(self, resource):
        self._queue.stash(resource)
        super().stash(resource)

    def recover(self, resource):
        self._queue.recover(resource)
        if len(self._queue) > self._num_of_spill:  # 为了兼容已经存在的 RecoverableRequestQueue，可以在一段时间后删除
            requests = deque()
            while not self._queue.empty():
                requests.append(self._queue.get())
            while requests:
                self.put(requests.pop())
        try:
            super().recover(resource)
        except:
            pass

    def can_recover(self, resource):
        return self._queue.can_recover(resource)


def get_queue_for_spider(spider):
    if not isinstance(spider, Spider):
        raise TypeError("unknown spider type `{}`".format(spider.__class__.__name__))
    num_of_spill = spider.num_of_spill
    if isinstance(spider, RecoverableSpider):
        queue = RecoverableRequestQueue()
        if num_of_spill:
            return RecoverableSpillRequestQueueProxy(queue, num_of_spill=num_of_spill)
        else:
            return queue
    if isinstance(spider, AsyncSpider):
        queue = SimpleRequestQueue()
        if num_of_spill:
            return SpillRequestQueueProxy(queue, num_of_spill=num_of_spill)
        else:
            return queue
